require "socket"

class MySql::Connection < DB::Connection
  record Options,
    host : String,
    port : Int32,
    username : String?,
    password : String?,
    initial_catalog : String?,
    charset : String do
    def self.from_uri(uri : URI) : Options
      host = uri.hostname || raise "no host provided"
      port = uri.port || 3306
      username = uri.user
      password = uri.password

      charset = uri.query_params.fetch "encoding", Collations.default_collation

      path = uri.path
      if path && path.size > 1
        initial_catalog = path[1..-1]
      else
        initial_catalog = nil
      end

      Options.new(
        host: host, port: port, username: username, password: password,
        initial_catalog: initial_catalog, charset: charset)
    end
  end

  def initialize(options : ::DB::Connection::Options, mysql_options : ::MySql::Connection::Options)
    super(options)
    @socket = uninitialized TCPSocket

    begin
      charset_id = Collations.id_for_collation(mysql_options.charset).to_u8

      @socket = TCPSocket.new(mysql_options.host, mysql_options.port)
      handshake = read_packet(Protocol::HandshakeV10)

      write_packet(1) do |packet|
        Protocol::HandshakeResponse41.new(mysql_options.username, mysql_options.password, mysql_options.initial_catalog, handshake.auth_plugin_data, charset_id).write(packet)
      end

      read_ok_or_err do |packet, status|
        raise "packet #{status} not implemented"
      end
    rescue IO::Error
      raise DB::ConnectionRefused.new
    end
  end

  def do_close
    super

    begin
      write_packet do |packet|
        Protocol::Quit.new.write(packet)
      end
      @socket.close
    rescue
    end
  end

  # :nodoc:
  def read_ok_or_err
    read_packet do |packet|
      raise_if_err_packet(packet) do |status|
        yield packet, status
      end
    end
  end

  # :nodoc:
  def read_packet
    packet = build_read_packet
    begin
      yield packet
    ensure
      packet.discard
    end
  end

  # :nodoc:
  def read_packet(protocol_packet_type)
    read_packet do |packet|
      return protocol_packet_type.read(packet)
    end
    raise "unable to read packet"
  end

  # :nodoc:
  def build_read_packet
    ReadPacket.new(@socket, self)
  end

  # :nodoc:
  def write_packet(seq = 0)
    content = IO::Memory.new
    yield WritePacket.new(content, self)
    bytesize = content.bytesize

    packet = IO::Memory.new
    3.times do
      packet.write_byte (bytesize & 0xff_u8).to_u8
      bytesize >>= 8
    end
    packet.write_byte seq.to_u8

    packet << content

    @socket << packet
    @socket.flush
  end

  # :nodoc:
  def handle_err_packet(packet)
    8.times { packet.read_byte! }
    raise packet.read_string
  end

  # :nodoc:
  def raise_if_err_packet(packet)
    raise_if_err_packet(packet) do |status|
      raise "unexpected packet #{status}"
    end
  end

  # :nodoc:
  def raise_if_err_packet(packet)
    status = packet.read_byte!
    if status == 255
      handle_err_packet packet
    end

    yield status if status != 0

    status
  end

  # :nodoc:
  def read_column_definitions(target, column_count)
    # Parse column definitions
    # http://dev.mysql.com/doc/internals/en/com-query-response.html#packet-Protocol::ColumnDefinition
    column_count.times do
      self.read_packet do |packet|
        catalog = packet.read_lenenc_string
        schema = packet.read_lenenc_string
        table = packet.read_lenenc_string
        org_table = packet.read_lenenc_string
        name = packet.read_lenenc_string
        org_name = packet.read_lenenc_string
        next_length = packet.read_lenenc_int # length of fixed-length fields, always 0x0c
        raise "Unexpected next_length value: #{next_length}." unless next_length == 0x0c
        character_set = packet.read_fixed_int(2).to_u16!
        column_length = packet.read_fixed_int(4).to_u32!
        column_type = packet.read_fixed_int(1).to_u8!
        flags = packet.read_fixed_int(2).to_u16!
        decimal = packet.read_fixed_int(1).to_u8!
        filler = packet.read_fixed_int(2).to_u16! # filler [00] [00]
        raise "Unexpected filler value #{filler}" unless filler == 0x0000

        target << ColumnSpec.new(catalog, schema, table, org_table, name, org_name, character_set, column_length, column_type, flags, decimal)
      end
    end

    if column_count > 0
      self.read_packet do |eof_packet|
        eof_packet.read_byte # TODO assert EOF Packet
      end
    end
  end

  def build_prepared_statement(query) : MySql::Statement
    MySql::Statement.new(self, query)
  end

  def build_unprepared_statement(query) : MySql::UnpreparedStatement
    MySql::UnpreparedStatement.new(self, query)
  end
end
